"""
Implements the "option groups" feature.
"""
from collections import defaultdict
from typing import (
    Callable, Dict, Iterable, Iterator, List, Optional, Sequence, Tuple, overload,
)

import click
from click import Option, Parameter

import cloup
from cloup._params import option
from cloup._util import first_bool, make_repr
from cloup.constraints import Constraint
from cloup.formatting import HelpSection, ensure_is_cloup_formatter
from cloup.typing import Decorator, F


class OptionGroup:
    def __init__(self, title: str,
                 help: Optional[str] = None,
                 constraint: Optional[Constraint] = None,
                 hidden: bool = False):
        """Contains the information of an option group and identifies it.
        Note that, as far as the clients of this library are concerned, an
        ``OptionGroups`` acts as a "marker" for options, not as a container for
        related options. When you call ``@optgroup.option(...)`` you are not
        adding an option to a container, you are just adding an option marked
        with this option group.

        .. versionadded:: 0.8.0
            The ``hidden`` parameter.
        """
        if not title:
            raise ValueError('name is a mandatory argument')  # pragma: no cover
        self.title = title
        self.help = help
        self._options: Sequence[click.Option] = []
        self.constraint = constraint
        self.hidden = hidden

    @property
    def options(self) -> Sequence[click.Option]:
        return self._options

    @options.setter
    def options(self, options: Iterable[click.Option]) -> None:
        self._options = opts = tuple(options)
        if self.hidden:
            for opt in opts:
                opt.hidden = True
        elif all(opt.hidden for opt in opts):
            self.hidden = True

    def get_help_records(self, ctx: click.Context) -> List[Tuple[str, str]]:
        if self.hidden:
            return []
        return [
            opt.get_help_record(ctx) for opt in self if not opt.hidden  # type: ignore
        ]  # get_help_record() should return None only if opt.hidden

    def option(self, *param_decls, **attrs) -> Callable[[F], F]:
        """Refer to :func:`cloup.option`."""
        return option(*param_decls, group=self, **attrs)

    def __iter__(self) -> Iterator[click.Option]:
        return iter(self.options)

    def __getitem__(self, i: int) -> click.Option:
        return self.options[i]

    def __len__(self) -> int:
        return len(self.options)

    def __repr__(self) -> str:
        return make_repr(self, self.title, help=self.help, options=self.options)

    def __str__(self) -> str:
        return make_repr(
            self, self.title, options=[opt.name for opt in self.options])


def has_option_group(param) -> bool:
    return getattr(param, 'group', None) is not None


def get_option_group_of(param) -> Optional[OptionGroup]:
    return getattr(param, 'group', None)


# noinspection PyMethodMayBeStatic
class OptionGroupMixin:
    """Implements support to:

     - option groups
     - the "Positional arguments" help section; this section is shown only if
       at least one of your arguments has non-empty ``help``.

    .. important::
        In order to check the constraints defined on the option groups,
        a command must inherits from :class:`cloup.ConstraintMixin` too!

    .. versionadded:: 0.14.0
        added the "Positional arguments" help section.

    .. versionchanged:: 0.8.0
        this mixin now relies on ``cloup.HelpFormatter`` to align help sections.
        If a ``click.HelpFormatter`` is used with a ``TypeError`` is raised.

    .. versionchanged:: 0.8.0
        removed ``format_option_group``. Added ``get_default_option_group`` and
        ``make_option_group_help_section``.

    .. versionadded:: 0.5.0
    """

    def __init__(
        self, *args, align_option_groups: Optional[bool] = None, **kwargs
    ):
        """
        :param align_option_groups:
            whether to align the columns of all option groups' help sections.
            This is also available as a context setting having a lower priority
            than this attribute. Given that this setting should be consistent
            across all you commands, you should probably use the context
            setting only.
        :param args:
            positional arguments forwarded to the next class in the MRO
        :param kwargs:
            keyword arguments forwarded to the next class in the MRO
        """
        self.align_option_groups = align_option_groups
        params = kwargs.get('params') or []
        arguments, option_groups, ungrouped_options = self._group_params(params)

        self.arguments = arguments

        self.option_groups = option_groups
        """List of all option groups, except the "default option group"."""

        self.ungrouped_options = ungrouped_options
        """List of options not explicitly assigned to an user-defined option group.
        These options will be included in the "default option group".
        **Note:** this list does not include options added automatically by Click
        based on context settings, like the ``--help`` option; use the
        :meth:`get_ungrouped_options` method if you need the real full list
        (which needs a ``Context`` object)."""

        super().__init__(*args, **kwargs)  # type: ignore

    @staticmethod
    def _group_params(
        params: List[Parameter]
    ) -> Tuple[List[click.Argument], List[OptionGroup], List[Option]]:

        options_by_group: Dict[OptionGroup, List[click.Option]] = defaultdict(list)
        arguments: List[click.Argument] = []
        ungrouped_options: List[click.Option] = []
        for param in params:
            if isinstance(param, click.Argument):
                arguments.append(param)
            elif isinstance(param, click.Option):
                grp = get_option_group_of(param)
                if grp is None:
                    ungrouped_options.append(param)
                else:
                    options_by_group[grp].append(param)

        option_groups = list(options_by_group.keys())
        for group, options in options_by_group.items():
            group.options = options

        return arguments, option_groups, ungrouped_options

    def get_ungrouped_options(self, ctx: click.Context) -> Sequence[click.Option]:
        """Returns options not explicitly assigned to an option group
        (eventually including the ``--help`` option), i.e. options that will be
        part of the "default option group"."""
        help_option = ctx.command.get_help_option(ctx)
        if help_option is not None:
            return self.ungrouped_options + [help_option]
        else:
            return self.ungrouped_options

    def get_argument_help_record(
        self, arg: click.Argument, ctx: click.Context
    ) -> Tuple[str, str]:
        if isinstance(arg, cloup.Argument):
            return arg.help_record(ctx)
        return arg.make_metavar(), ""

    def get_arguments_help_section(self, ctx: click.Context) -> Optional[HelpSection]:
        if not any(getattr(arg, "help", None) for arg in self.arguments):
            return None
        return HelpSection(
            heading="Positional arguments",
            definitions=[
                self.get_argument_help_record(arg, ctx) for arg in self.arguments
            ]
        )

    def make_option_group_help_section(
        self, group: OptionGroup, ctx: click.Context
    ) -> HelpSection:
        """Returns a HelpSection for an OptionGroup, i.e. an object containing
        the title, the optional description and the options' definitions for
        this option group.

        .. versionadded:: 0.8.0
        """
        return HelpSection(
            heading=group.title,
            definitions=group.get_help_records(ctx),
            help=group.help,
            constraint=group.constraint.help(ctx) if group.constraint else None
        )

    def must_align_option_groups(
        self, ctx: Optional[click.Context], default: bool = True
    ) -> bool:
        """
        Returns ``True`` if the help sections of all options groups should have
        their columns aligned.

        .. versionadded:: 0.8.0
        """
        return first_bool(
            self.align_option_groups,
            getattr(ctx, 'align_option_groups', None),
            default,
        )

    def get_default_option_group(
        self, ctx: click.Context, is_the_only_option_group: bool = False
    ) -> OptionGroup:
        """
        Returns an ``OptionGroup`` instance for the options not explicitly
        assigned to an option group, eventually including the ``--help`` option.

        .. versionadded:: 0.8.0
        """
        default_group = OptionGroup(
            "Options" if is_the_only_option_group else "Other options")
        default_group.options = self.get_ungrouped_options(ctx)
        return default_group

    def format_params(
        self, ctx: click.Context, formatter: click.HelpFormatter
    ) -> None:
        formatter = ensure_is_cloup_formatter(formatter)

        visible_sections = []
        positional_arguments_section = self.get_arguments_help_section(ctx)
        if positional_arguments_section:
            visible_sections.append(positional_arguments_section)
        visible_sections += [
            self.make_option_group_help_section(group, ctx)
            for group in self.option_groups
            if not group.hidden
        ]

        default_group = self.get_default_option_group(
            ctx, is_the_only_option_group=not visible_sections
        )
        if not default_group.hidden:
            visible_sections.append(
                self.make_option_group_help_section(default_group, ctx))

        formatter.write_many_sections(
            visible_sections,
            aligned=self.must_align_option_groups(ctx),
        )


@overload
def option_group(
    title: str,
    help: str,
    *options: Decorator,
    constraint: Optional[Constraint] = None,
    hidden: bool = False,
) -> Callable[[F], F]:
    ...


@overload
def option_group(
    title: str,
    *options: Decorator,
    help: Optional[str] = None,
    constraint: Optional[Constraint] = None,
    hidden: bool = False,
) -> Callable[[F], F]:
    ...


# noinspection PyIncorrectDocstring
def option_group(title, *args, **kwargs):
    """
    Returns a decorator that annotates a function with an option group.

    The ``help`` argument is an optional description and can be provided either
    as keyword argument or as 2nd positional argument after the ``name`` of
    the group::

        # help as keyword argument
        @option_group(name, *options, help=None, ...)

        # help as 2nd positional argument
        @option_group(name, help, *options, ...)

    .. versionchanged:: 0.9.0
        in order to support the decorator :func:`cloup.constrained_params`,
        ``@option_group`` now allows each input decorators to add multiple
        options.

    :param title:
        title of the help section describing the option group.
    :param help:
        an optional description shown below the name; can be provided as keyword
        argument or 2nd positional argument.
    :param options:
        an arbitrary number of decorators like ``click.option``, which attach
        one or multiple options to the decorated command function.
    :param constraint:
        an optional instance of :class:`~cloup.constraints.Constraint`
        (see :doc:`Constraints </pages/constraints>` for more info);
        a description of the constraint will be shown between squared brackets
        aside the option group title (or below it if too long).
    :param hidden:
        if ``True``, the option group and all its options are hidden from the help page
        (all contained options will have their ``hidden`` attribute set to ``True``).
    """
    if args and isinstance(args[0], str):
        return _option_group(title, options=args[1:], help=args[0], **kwargs)
    else:
        return _option_group(title, options=args, **kwargs)


def _option_group(
    title: str,
    options: Callable[[F], F],
    help: Optional[str] = None,
    constraint: Optional[Constraint] = None,
    hidden: bool = False,
) -> Callable[[F], F]:
    if not isinstance(title, str):
        raise TypeError(
            'the first argument of @option_group must be its title, a string. '
            'You probably forgot it!'
        )

    if not options:
        raise ValueError('you must provide at least one option')

    def decorator(f):
        opt_group = OptionGroup(title, help=help, constraint=constraint, hidden=hidden)
        if not hasattr(f, '__click_params__'):
            f.__click_params__ = []
        for add_option in reversed(options):
            prev_len = len(f.__click_params__)
            add_option(f)
            added_options = f.__click_params__[prev_len:]
            for new_option in added_options:
                if not isinstance(new_option, Option):
                    raise TypeError(
                        "only parameter of type `Option` can be added to option groups")
                if has_option_group(new_option):
                    raise ValueError(
                        f'{new_option} was first assigned to {new_option.group} and then '
                        f'passed as argument to @option_group({title!r}, ...)'
                    )
                new_option.group = opt_group
                if hidden:
                    new_option.hidden = True
        return f

    return decorator



"""
This example shows how to use cloup.Section to organize the subcommands of
a multi-command in many --help sections.

The code was generated by parsing "git --help" and taking
- the first 3 sections
- for each section, the first 3 commands after shuffling them.
"""
# flake8: noqa E128
import cloup


def f(**kwargs):
    """ Dummy command callback """
    print(**kwargs)


# In a real big application, you would import the following commands from separate modules =========
git_clone = cloup.command('clone', help='Clone a repository into a new directory')(f)
git_init = cloup.command('init', help='Create an empty Git repository or reinitialize an existing one')(f)

git_rm = cloup.command('rm', help='Remove files from the working tree and from the index')(f)
git_sparse_checkout = cloup.command('sparse-checkout', help='Initialize and modify the sparse-checkout')(f)
git_mv = cloup.command('mv', help='Move or rename a file, a directory, or a symlink')(f)

git_status = cloup.command('status', help='Show the working tree status')(f)
git_diff = cloup.command('diff', help='Show changes between commits, commit and working tree, etc')(f)
git_bisect = cloup.command('bisect', help='Use binary search to find the commit that introduced a bug')(f)


# ==================================================================================================


# If "align_sections=True" (default), the help column of all sections will
# be aligned; otherwise, each section will be formatted independently.
@cloup.group('git', align_sections=True)
def git():
    return 0


"""
git.section() creates a new Section object, adds it to git and returns it.

In the help, sections are shown in the same order they are added.
Commands in each sections are shown in the same order they are listed, unless
you pass the argument "sorted=True".
"""
git.section(
    'Start a working area (see also: git help tutorial)',
    git_clone,
    git_init,
)
git.section(
    'Work on the current change (see also: git help everyday)',
    git_rm,
    git_sparse_checkout,
    git_mv,
)
git.section(
    'Examine the history and state (see also: git help revisions)',
    git_status,
    git_diff,
    git_bisect,
)

"""
In alternative, you can either:
- pass a list of Section objects as argument "sections" to cloup.Group or @cloup.group
- use git.add_section(section) to add an existing Section object
- use git.add_command(cmd, name, section, ...); the section must NOT contain the command
- use @git.command(cmd, name, section, ...)

Individual commands don't store the section they belong to.
Also, cloup.command doesn't accept a "section" argument.
"""

# The following commands will be added to the "default section" (a sorted Section)
git.add_command(cloup.command('fake-2', help='Fake command #2')(f))
git.add_command(cloup.command('fake-1', help='Fake command #1')(f))

if __name__ == '__main__':
    git()



"""
This module contains anything related to separators. Currently, it contains an
implementation of "row sep policies", e.g. components that decide if and how to
add an extra separator/spacing between the rows of a definition list (only for
tabular layout). In the future, it may be expanded with something analogous for
help sections.
"""
import abc
import sys
from itertools import zip_longest
from typing import Optional, Sequence, Union

if sys.version_info[:2] >= (3, 8):
    from typing import Protocol
else:  # pragma: no cover
    from typing_extensions import Protocol


SepType = Union[str, 'SepGenerator']


class SepGenerator(Protocol):
    """Generates a separator given a width. When used as ``row_sep``, this ``width`
    corresponds to ``HelpFormatter.available_width``, i.e. the line width excluding
    the current indentation width.

    Note: the length of the returned separator may differ from ``width``.
    """
    def __call__(self, width: int) -> str:
        ...


class RowSepPolicy(metaclass=abc.ABCMeta):
    """A callable that can be passed as ``row_sep`` to  :class:`HelpFormatter` in
    order to decide *if* a definition list should get a row separator (in
    addition to ``\\n``) and *which* separator.

    In practice, the row separator should be the same for all definition lists
    that satisfy a given condition. That's why :class:`RowSepIf`
    exists, you probably want to use that.

    Nonetheless, this protocol exists mainly for one reason: it leaves open the
    door to policies that can decide a row separator for each individual row
    (feature I'm personally against to for now), without breaking changes.
    This would make possible to implement the old Click 7.2 behavior, which
    inserted an empty line only after options with a long help. Adding this
    feature would be possible without breaking changes, by extending the return
    type to ``Union[None, str, Sequence[str]]``.
    """

    @abc.abstractmethod
    def __call__(  # noqa E704
        self, rows: Sequence[Sequence[str]],
        col_widths: Sequence[int],
        col_spacing: int,
    ) -> Optional[str]:
        """Decides which row separator to use (eventually none) in the given
        definition list."""


class RowSepCondition(Protocol):
    """Determines when a definition list should use a row separator."""

    # Ignore error due to flake8 issue: "multiple statements on one line (def)"
    def __call__(  # noqa E704
        self, rows: Sequence[Sequence[str]],
        col_widths: Sequence[int],
        col_spacing: int,
    ) -> bool:
        """Returns ``True`` if the input definition list should use a row
        separator (in addition to the usual ``\\n``)."""


class RowSepIf(RowSepPolicy):
    def __init__(self, condition: RowSepCondition,
                 sep: Union[str, SepGenerator] = ''):
        """
        Inserts a row separator between the rows of a definition list only if a
        condition is satisfied. This class implements the ``RowSepPolicy``
        protocol and does two things:

        - enforces the use of a single row separator for all rows of a
          definition lists and for all definition lists; note that
          ``RowSepPolicy`` doesn't for implementation reasons but it's probably
          what you want;
        - allows you to implement different conditions (see type
          :data:`RowSepCondition`) without worrying about the generation part,
          which is always the same.

        :param condition:
            a :class:`RowSepCondition` that determines when to add the (extra)
            row separator.
        :param sep:
            either a string or a ``SepGenerator``,
            i.e. a function ``(width: int) -> str`` (e.g. :class:`Hline`).
            The empty string corresponds to an empty line separator.
        """
        if isinstance(sep, str) and sep.endswith('\n'):
            raise ValueError(
                "sep must not end with '\\n'. The formatter writes  a '\\n' after it; "
                "no other newline is allowed.")
        self.condition = condition
        self.sep = sep

    def __call__(
        self, rows: Sequence[Sequence[str]], col_widths: Sequence[int], col_spacing: int
    ) -> Optional[str]:
        if self.condition(rows, col_widths, col_spacing):
            if callable(self.sep):
                total_width = get_total_width(col_widths, col_spacing)
                return self.sep(total_width)
            return self.sep
        return None


# ==========================================
#  Conditions & related utils

def get_total_width(col_widths: Sequence[int], col_spacing: int) -> int:
    """Returns the total width of a definition list (or, more generally, a table).
    Useful when implementing a RowSepStrategy."""
    return sum(col_widths) + col_spacing * (len(col_widths) - 1)


def count_multiline_rows(rows: Sequence[str], col_widths: Sequence[int]) -> int:
    # Note: I'm using zip_longest on purpose so that a TypeError will be raised
    # if len(row) != len(col_widths). An explicit check is not worth it since
    # this should never happen.
    return sum(
        any(len(col_text) > col_width
            for col_text, col_width in zip_longest(row, col_widths))
        for row in rows
    )


def multiline_rows_are_at_least(
    count_or_percentage: Union[int, float]
) -> RowSepCondition:
    """
    Returns a ``RowSepStrategy`` that returns a row separator between all rows
    of a definition list, only if the number of rows taking multiple lines is
    greater than or equal to a certain threshold.

    :param count_or_percentage:
        a threshold for multiline rows above which the returned strategy will
        insert a row separator. It can be either an absolute count (`int`) or a
        percentage relative to the total number of rows expressed as a `float`
        between 0 and 1 (0 and 1 excluded).
    """
    if count_or_percentage <= 0:
        raise ValueError('count_or_percentage should be > 0')

    if isinstance(count_or_percentage, int):
        count_threshold = count_or_percentage

        def condition(rows, col_widths, col_spacing):
            num_multiline = count_multiline_rows(rows, col_widths)
            return num_multiline >= count_threshold

    elif isinstance(count_or_percentage, float):
        percent_threshold = count_or_percentage
        if percent_threshold > 1.0:
            raise ValueError(
                "count_or_percentage must be either an integer or a float in the "
                f"interval ]0, 1[. You passed a float >= 1.0 ({percent_threshold}).")

        def condition(rows, col_widths, col_spacing):
            num_multiline = count_multiline_rows(rows, col_widths)
            percent_multiline = num_multiline / len(rows)
            return percent_multiline >= percent_threshold
    else:
        raise TypeError('count_or_percentage must be an int or a float')

    return condition


class Hline(SepGenerator):
    # Workaround: PyCharm auto-completion doesn't work without these declarations
    solid: 'Hline'
    dashed: 'Hline'
    densely_dashed: 'Hline'
    dotted: 'Hline'

    def __init__(self, pattern: str):
        """Returns a function that generates an horizontal line of a given length.

        This class has different static members for different line styles
        like ``Hline.solid``, ``Hline.dashed``, ``Hline.densely_dashed``
        and  ``Hline.dotted``.

        :param pattern:
            a string (usually a single character) that is repeated to generate
            the line.
        """
        self.pattern = pattern

    def __call__(self, width: int) -> str:
        pattern = self.pattern
        if len(pattern) == 1:
            return pattern * width
        reps, rest = width // len(pattern), width % len(pattern)
        return pattern * reps + pattern[:rest]


Hline.solid = Hline("─")
"""Returns a line like ``────────``."""

Hline.dashed = Hline('-')
"""Returns a line like ``--------``."""

Hline.densely_dashed = Hline('╌')
"""Returns a line like ``╌╌╌╌╌╌╌╌``."""

Hline.dotted = Hline("┄")
"""Returns a line like ``┄┄┄┄┄┄┄┄``."""



from contextlib import contextmanager
from typing import Iterable, List
from unittest.mock import Mock

import click
import pytest

import cloup
from cloup import Context


def pick_first_bool(args: Iterable, *, default: bool) -> bool:
    return next((arg for arg in args if isinstance(arg, bool)), default)


def new_dummy_func():
    return lambda *args, **kwargs: 1


def int_opt(*args, **kwargs):
    return click.Option(*args, type=int, **kwargs)


def bool_opt(*args, **kwargs):
    return click.Option(*args, type=bool, **kwargs)


def flag_opt(*args, **kwargs):
    return click.Option(*args, is_flag=True, **kwargs)


def multi_opt(*args, **kwargs):
    return click.Option(*args, multiple=True, **kwargs)


def tuple_opt(*args, **kwargs):
    return click.Option(*args, nargs=3, **kwargs)


def parametrize(argnames, *argvalues, **kwargs):
    return pytest.mark.parametrize(argnames, argvalues, **kwargs)


def make_context(cmd: click.Command, shell: str) -> click.Context:
    args = shell.split()
    return cmd.make_context(cmd.name, args)


def make_fake_context(
    params: Iterable[click.Parameter],
    command_cls=cloup.Command,
    cls=Context,
    **ctx_kwargs
) -> Context:
    """Creates a simple instance of Command with the specified parameters,
    then create a fake context without actually invoking the command."""
    return cls(
        command_cls('fake', params=params, callback=new_dummy_func()), **ctx_kwargs
    )


def make_options(names: Iterable[str], **common_kwargs) -> List[click.Option]:
    return [click.Option([f'--{name}'], **common_kwargs) for name in names]


def should_raise(expected_exception, *, when, **kwargs):
    if when:
        return pytest.raises(expected_exception, **kwargs)

    @contextmanager
    def manager():
        yield

    return manager()


def mock_repr(value, *args, **kwargs):
    m = Mock(*args, **kwargs)
    m.__repr__ = lambda x: value
    return m



"""
Types for parameter decorators are in this stub for convenience of implementation.
"""
from typing import Any, Callable, Optional, Sequence, Tuple, Type, TypeVar, Union

import click

from cloup import OptionGroup

F = TypeVar('F', bound=Callable)
P = TypeVar('P', bound=click.Parameter)

ParamTypeLike = Union[
    click.ParamType,
    Type[float], Type[int],
    Type[str], Type[tuple],
]
ParamDefault = Union[Any, Callable[[], Any]]
ParamCallback = Callable[[click.Context, P, Any], Any]

# Click 8 deprecates the argument `autocompletion` and replaces it with the
# argument `shell_complete`, which has a different semantic.
# The following will be uncommented when Cloup drop support for Click 7:
#
# ShellComplete = Callable[
#     [click.Context, P, str],
#     Union[List['CompletionItem'], List[str]],
# ]

class Argument(click.Argument):
    def __init__(self, *args, help: Optional[str] = None, **attrs):
        ...

    def help_record(self, ctx: click.Context) -> Tuple[str, str]:
        ...


class Option(click.Option):
    def __init__(self, *args, group: Optional[OptionGroup] = None, **attrs):
        ...


def argument(
    *param_decls,
    cls: Optional[Type[click.Argument]] = None,
    type: Optional[ParamTypeLike] = None,
    required: Optional[bool] = None,
    default: Optional[ParamDefault] = None,
    callback: Optional[ParamCallback[click.Argument]] = None,
    nargs: Optional[int] = None,
    metavar: Optional[str] = None,
    expose_value: bool = True,
    envvar: Optional[Union[str, Sequence[str]]] = None,
    # The following will be added when Cloup drops support for Click 7:
    #     shell_complete: Optional[ShellComplete] = None,
    **kwargs
) -> Callable[[F], F]: ...


def option(
    *param_decls,
    cls: Optional[Type[click.Option]] = None,
    # Commonly used
    metavar: Optional[str] = None,
    type: Optional[ParamTypeLike] = None,
    is_flag: Optional[bool] = None,
    default: Optional[ParamDefault] = None,
    required: Optional[bool] = None,
    help: Optional[str] = None,
    # Processing
    callback: Optional[ParamCallback[click.Option]] = None,
    is_eager: bool = False,
    # Help text tuning
    show_choices: bool = True,
    show_default: bool = False,
    show_envvar: bool = False,
    # Flag options
    flag_value: Optional[Any] = None,
    count: bool = False,
    # Multiple values
    nargs: Optional[int] = None,
    multiple: bool = False,
    # Prompt
    prompt: Union[bool, str] = False,
    confirmation_prompt: Union[bool, str] = False,
    prompt_required: bool = True,
    hide_input: bool = False,
    # Environment
    allow_from_autoenv: bool = True,
    envvar: Optional[Union[str, Sequence[str]]] = None,
    # Hiding
    hidden: bool = False,
    expose_value: bool = True,
    # Others
    group: Optional[OptionGroup] = None,
    # The following will be added when Cloup drops support for Click 7:
    #     shell_complete: Optional[ShellComplete] = None,
    **kwargs
) -> Callable[[F], F]: ...



"""
Example based on the CLI of Manim Community, which actually uses Cloup.
This example shows how a real-world application could look like and serves to me
as a test bench for trying out styling and formatting.
"""
import cloup
from cloup import (
    Color,
    Context,
    HelpFormatter,
    HelpTheme,
    Style,
)
from cloup.formatting.sep import RowSepIf, multiline_rows_are_at_least
from config import config
from render import render

VERSION = '0.5.0'
CONTEXT_SETTINGS = Context.settings(
    help_option_names=["-h", "--help"],
    align_option_groups=False,
    align_sections=True,  # subcommand help sections
    # color=False,
    show_subcommand_aliases=True,
    formatter_settings=HelpFormatter.settings(
        # width=None,
        # max_width=80,
        # col1_max_width=30
        # col2_min_width=35,
        # indent_increment=2,
        row_sep=RowSepIf(
            multiline_rows_are_at_least(0.35),
            # sep=Hline.densely_dashed
        ),
        theme=HelpTheme.dark().with_(
            # invoked_command=Style(...),
            # command_help=Style(...),
            # heading=Style(...),
            # constraint=Style(fg='red'),
            # section_help=Style(...),
            # col1=Style(...),
            col2=Style(dim=True),
            epilog=Style(fg=Color.bright_white, italic=True),
        ),
    ),
)


@cloup.group(
    name='Manim',
    no_args_is_help=True,
    context_settings=CONTEXT_SETTINGS,
    epilog="Made with <3 by Manim Community developers.",
)
@cloup.version_option(version=VERSION)
def main():
    """Animation engine for explanatory math videos."""
    pass


main.add_command(render)
main.add_command(config)

if __name__ == "__main__":
    main(prog_name='manim')



import abc
from typing import (
    Callable, Iterable, Optional, Sequence, TypeVar, Union, overload,
)

import click
from click import Context, Parameter

from cloup._util import (
    FrozenSpace, check_arg, class_name,
    first_bool, make_one_line_repr, make_repr, pluralize, reindent,
)
from .common import (
    format_param_list,
    get_param_label,
    get_param_name,
    get_params_whose_value_is_set,
    get_required_params,
    param_value_is_set,
)
from .exceptions import ConstraintViolated, UnsatisfiableConstraint
from ..typing import Decorator, F

Op = TypeVar('Op', bound='Operator')
HelpRephraser = Callable[[Context, 'Constraint'], str]
ErrorRephraser = Callable[[ConstraintViolated], str]


class Constraint(abc.ABC):
    """
    A constraint that can be checked against an arbitrary collection of CLI
    parameters with respect to a specific :class:`click.Context` (which
    contains the values assigned to the parameters in ``ctx.params``).

    .. versionchanged:: 0.9.0
        calling a constraint, previously equivalent to :meth:`~Constraint.check`,
        is now equivalent to calling :func:`cloup.constrained_params` with this
        constraint as first argument.
    """

    @staticmethod
    def must_check_consistency(ctx: click.Context) -> bool:
        """Returns True if consistency checks are enabled.

        .. versionchanged:: 0.9.0
            this method now a static method and takes a ``Context`` in input.
        """
        return first_bool(
            getattr(ctx, 'check_constraints_consistency', True),
            True,
        )

    def __getattr__(self, attr):
        removed_attrs = ('toggle_consistency_checks', 'consistency_checks_toggled')
        if attr in removed_attrs:
            raise Exception(
                f'{attr} was removed in v0.9.0. You can now enable/disable consistency'
                f'checks using the Context parameter check_constraints_consistency. '
                f'Pass it as part of your context_settings.'
            )

    @abc.abstractmethod
    def help(self, ctx: Context) -> str:
        """A description of the constraint. """

    def check_consistency(self, params: Sequence[Parameter]) -> None:
        """
        Performs some sanity checks that detect inconsistencies between this
        constraints and the properties of the input parameters (e.g. required).

        For example, a constraint that requires the parameters to be mutually
        exclusive is not consistent with a group of parameters with multiple
        required options.

        These sanity checks are meant to catch developer's mistakes and don't
        depend on the values assigned to the parameters; therefore:

        - they can be performed before any parameter parsing
        - they can be disabled in production (setting
          ``check_constraints_consistency=False`` in ``context_settings``)

        :param params: list of :class:`click.Parameter` instances
        :raises: :exc:`~cloup.constraints.errors.UnsatisfiableConstraint`
                 if the constraint cannot be satisfied independently from the values
                 provided by the user
        """

    @abc.abstractmethod
    def check_values(self, params: Sequence[Parameter], ctx: Context):
        """
        Checks that the constraint is satisfied by the input parameters in the
        given context, which (among other things) contains the values assigned
        to the parameters in ``ctx.params``.

        You probably don't want to call this method directly.
        Use :meth:`check` instead.

        :param params: list of :class:`click.Parameter` instances
        :param ctx: :class:`click.Context`
        :raises:
            :exc:`~cloup.constraints.ConstraintViolated`
        """

    @overload
    def check(self, params: Sequence[Parameter], ctx: Optional[Context] = None) -> None:
        ...

    @overload
    def check(self, params: Iterable[str], ctx: Optional[Context] = None) -> None:
        ...

    def check(self, params, ctx: Optional[Context] = None) -> None:
        """
        Raises an exception if the constraint is not satisfied by the input
        parameters in the given (or current) context.

        This method calls both :meth:`check_consistency` (if enabled) and
        :meth:`check_values`.

        .. tip::
            By default :meth:`check_consistency` is called since it shouldn't
            have any performance impact. Nonetheless, you can disable it in
            production passing ``check_constraints_consistency=False`` as part
            of your ``context_settings``.

        :param params: an iterable of parameter names or a sequence of
                       :class:`click.Parameter`
        :param ctx: a `Context`; if not provided, :func:`click.get_current_context`
                    is used
        :raises:
            :exc:`~cloup.constraints.ConstraintViolated`
            :exc:`~cloup.constraints.UnsatisfiableConstraint`
        """
        from ._support import ConstraintMixin

        if not params:
            raise ValueError("arg 'params' can't be empty")

        ctx = click.get_current_context() if ctx is None else ctx
        if not isinstance(ctx.command, ConstraintMixin):  # this is needed for mypy
            raise TypeError('constraints work only if the command inherits from '
                            'ConstraintMixin')

        params_objects = (ctx.command.get_params_by_name(params)
                          if isinstance(params[0], str)
                          else params)
        if Constraint.must_check_consistency(ctx):
            self.check_consistency(params_objects)
        return self.check_values(params_objects, ctx)

    def rephrased(
        self,
        help: Union[None, str, HelpRephraser] = None,
        error: Union[None, str, ErrorRephraser] = None,
    ) -> 'Rephraser':
        """
        Overrides the help string and/or the error message of this constraint
        wrapping it with a :class:`Rephraser`.

        :param help:
            if provided, overrides the help string of this constraint. It can be
            a string or a function ``(ctx: Context, constr: Constraint) -> str``.
            If you want to hide this constraint from the help, pass ``help=""``.
        :param error:
            if provided, overrides the error message of this constraint.
            It can be:

            - a string, eventually a ``format`` string supporting the replacement
              fields described in :class:`ErrorFmt`.

            - or a function ``(err: ConstraintViolated) -> str``; note that
              a :class:`ConstraintViolated` error has fields for ``ctx``,
              ``constraint`` and ``params``, so it's a complete description
              of what happened.
        """
        return Rephraser(self, help=help, error=error)

    def hidden(self) -> 'Rephraser':
        """Hides this constraint from the command help."""
        return Rephraser(self, help='')

    def __call__(self, *param_adders: Decorator) -> Callable[[F], F]:
        """Equivalent to calling :func:`cloup.constrained_params` with this
        constraint as first argument.

        .. versionchanged:: 0.9.0
            this method, previously equivalent to :meth:`~Constraint.check`, is
            now equivalent to calling :func:`cloup.constrained_params` with this
            constraint as first argument.
        """
        from ._support import constrained_params
        # TODO: remove this check in the future
        if not callable(param_adders[0]):
            from cloup import __version__
            raise TypeError(reindent(f"""\n
                since Cloup v0.9, calling a constraint has a completely different
                semantics and takes parameter decorators as arguments, see:

                https://cloup.readthedocs.io/en/v{__version__}/pages/constraints.html#constraints-as-decorators

                To check a constraint imperatively, you can use the check() method.
            """, 4))
        return constrained_params(self, *param_adders)

    def __or__(self, other: 'Constraint') -> 'Or':
        return Or(self, other)

    def __and__(self, other: 'Constraint') -> 'And':
        return And(self, other)

    def __repr__(self):
        return f'{class_name(self)}()'


class Operator(Constraint, abc.ABC):
    """Base class for all n-ary operators defined on constraints. """

    HELP_SEP: str
    """Used as separator of all constraints' help strings."""

    def __init__(self, *constraints: Constraint):
        """N-ary operator for constraints.
        :param constraints: operands
        """
        self.constraints = constraints

    def help(self, ctx: Context) -> str:
        return self.HELP_SEP.join(
            '(%s)' % c.help(ctx) if isinstance(c, Operator) else c.help(ctx)
            for c in self.constraints
        )

    def check_consistency(self, params: Sequence[Parameter]) -> None:
        for c in self.constraints:
            c.check_consistency(params)

    def __repr__(self):
        return make_repr(self, *self.constraints)


class And(Operator):
    """It's satisfied if all operands are satisfied."""
    HELP_SEP = ' and '

    def check_values(self, params: Sequence[Parameter], ctx: Context):
        for c in self.constraints:
            c.check_values(params, ctx)

    def __and__(self, other) -> 'And':
        if isinstance(other, And):
            return And(*self.constraints, *other.constraints)
        return And(*self.constraints, other)


class Or(Operator):
    """It's satisfied if at least one of the operands is satisfied."""
    HELP_SEP = ' or '

    def check_values(self, params: Sequence[Parameter], ctx: Context):
        for c in self.constraints:
            try:
                return c.check_values(params, ctx)
            except ConstraintViolated:
                pass
        raise ConstraintViolated.default(
            self.help(ctx), ctx=ctx, constraint=self, params=params
        )

    def __or__(self, other) -> 'Or':
        if isinstance(other, Or):
            return Or(*self.constraints, *other.constraints)
        return Or(*self.constraints, other)


class ErrorFmt(FrozenSpace):
    """:class:`Rephraser` allows you to pass a ``format`` string as ``error``
    argument; this class contains the "replacement fields" supported by such
    format string. You can use them as following::

        mutually_exclusive.rephrased(
            error=f"{ErrorFmt.error}\\n"
                  f"Some extra information here."
        )
    """

    error = '{error}'
    """Replaced by the original error message. Useful if all you want is to
    append or prepend some extra info to the original error message."""

    param_list = '{param_list}'
    """Replaced by a 2-space indented list of the constrained parameters."""


class Rephraser(Constraint):
    """A Constraint decorator that can override the help and/or the error
    message of the wrapped constraint.

    You'll rarely (if ever) use this class directly. In most cases, you'll use
    the method :meth:`Constraint.rephrased`. Refer to it for more info.

    .. seealso::

        - :meth:`Constraint.rephrased` -- wraps a constraint with a ``Rephraser``.
        - :class:`WrapperConstraint` -- alternative to ``Rephraser``.
        - :class:`ErrorFmt` -- describes the keyword you can use in an error
          format string.
    """

    def __init__(
        self, constraint: Constraint,
        help: Union[None, str, HelpRephraser] = None,
        error: Union[None, str, ErrorRephraser] = None,
    ):
        if help is None and error is None:
            raise ValueError('at least one between [help] and [error] must not be None')
        self.constraint = constraint
        self._help = help
        self._error = error

    def help(self, ctx: Context) -> str:
        if self._help is None:
            return self.constraint.help(ctx)
        elif isinstance(self._help, str):
            return self._help
        else:
            return self._help(ctx, self.constraint)

    def _get_rephrased_error(self, err: ConstraintViolated) -> Optional[str]:
        if self._error is None:
            return None
        elif isinstance(self._error, str):
            return self._error.format(
                error=str(err),
                param_list=format_param_list(err.params),
            )
        else:
            return self._error(err)

    def check_consistency(self, params: Sequence[Parameter]) -> None:
        try:
            self.constraint.check_consistency(params)
        except UnsatisfiableConstraint as exc:
            raise UnsatisfiableConstraint(
                self, params=params, reason=exc.reason)

    def check_values(self, params: Sequence[Parameter], ctx: Context):
        try:
            return self.constraint.check_values(params, ctx)
        except ConstraintViolated as err:
            rephrased_error = self._get_rephrased_error(err)
            if rephrased_error:
                raise ConstraintViolated(
                    rephrased_error, ctx=ctx, constraint=self, params=params)
            raise

    def __repr__(self):
        return make_one_line_repr(self, help=self._help)


class WrapperConstraint(Constraint, metaclass=abc.ABCMeta):
    """Abstract class that wraps another constraint and delegates all methods
    to it. Useful when you want to define a parametric constraint combining
    other existing constraints minimizing the boilerplate.

    This is an alternative to defining a function and using :class:`Rephraser`.
    Feel free to do that in your code, but cloup will stick to the convention
    that parametric constraints are defined as classes and written in
    camel-case."""

    def __init__(self, constraint: Constraint, **attrs):
        """
        :param constraint: the constraint to wrap
        :param attrs: these are just used to generate a ``__repr__`` method
        """
        self._constraint = constraint
        self._attrs = attrs

    def help(self, ctx: Context) -> str:
        return self._constraint.help(ctx)

    def check_consistency(self, params: Sequence[Parameter]) -> None:
        try:
            self._constraint.check_consistency(params)
        except UnsatisfiableConstraint as exc:
            raise UnsatisfiableConstraint(self, params=params, reason=exc.reason)

    def check_values(self, params: Sequence[Parameter], ctx: Context):
        self._constraint.check_values(params, ctx)

    def __repr__(self):
        return make_repr(self, **self._attrs)


class _RequireAll(Constraint):
    """Satisfied if all parameters are set."""

    def help(self, ctx: Context) -> str:
        return 'all required'

    def check_values(self, params: Sequence[Parameter], ctx: Context):
        values = ctx.params
        unset_params = [param for param in params
                        if not param_value_is_set(param, values[get_param_name(param)])]
        if any(unset_params):
            raise ConstraintViolated(
                pluralize(
                    len(unset_params),
                    one=f"{get_param_label(unset_params[0])} is required",
                    many=f"the following parameters are required:\n"
                         f"{format_param_list(unset_params)}"),
                ctx=ctx,
                constraint=self,
                params=params,
            )


class RequireAtLeast(Constraint):
    """Satisfied if the number of set parameters is >= n."""

    def __init__(self, n: int):
        check_arg(n >= 0)
        self.min_num_params = n

    def help(self, ctx: Context) -> str:
        return f'at least {self.min_num_params} required'

    def check_consistency(self, params: Sequence[Parameter]) -> None:
        n = self.min_num_params
        if len(params) < n:
            reason = (
                f'the constraint requires a minimum of {n} parameters but '
                f'it is applied on a group of only {len(params)} parameters!'
            )
            raise UnsatisfiableConstraint(self, params, reason)

    def check_values(self, params: Sequence[Parameter], ctx: Context):
        n = self.min_num_params
        given_params = get_params_whose_value_is_set(params, ctx.params)
        if len(given_params) < n:
            raise ConstraintViolated(
                f"at least {n} of the following parameters must be set:\n"
                f"{format_param_list(params)}",
                ctx=ctx, constraint=self, params=params,
            )

    def __repr__(self):
        return make_repr(self, self.min_num_params)


class AcceptAtMost(Constraint):
    """Satisfied if the number of set parameters is <= n."""

    def __init__(self, n: int):
        check_arg(n >= 0)
        self.max_num_params = n

    def help(self, ctx: Context) -> str:
        return f'at most {self.max_num_params} accepted'

    def check_consistency(self, params: Sequence[Parameter]) -> None:
        num_required_params = len(get_required_params(params))
        if num_required_params > self.max_num_params:
            reason = f'{num_required_params} of the parameters are required'
            raise UnsatisfiableConstraint(self, params, reason)

    def check_values(self, params: Sequence[Parameter], ctx: Context):
        n = self.max_num_params
        given_params = get_params_whose_value_is_set(params, ctx.params)
        if len(given_params) > n:
            raise ConstraintViolated(
                f"no more than {n} of the following parameters can be set:\n"
                f"{format_param_list(params)}",
                ctx=ctx, constraint=self, params=params,
            )

    def __repr__(self):
        return make_repr(self, self.max_num_params)


class RequireExactly(WrapperConstraint):
    """Requires an exact number of parameters to be set."""

    def __init__(self, n: int):
        check_arg(n > 0)
        self.num_params = n
        # Defined as a wrapper to reuse check_consistency() of the wrapped constraint.
        super().__init__(RequireAtLeast(n) & AcceptAtMost(n))

    def help(self, ctx: Context) -> str:
        return f'exactly {self.num_params} required'

    def check_values(self, params: Sequence[Parameter], ctx: Context):
        n = self.num_params
        given_params = get_params_whose_value_is_set(params, ctx.params)
        if len(given_params) != n:
            reason = pluralize(
                count=n,
                zero='none of the following parameters must be set:\n',
                many=f'exactly {n} of the following parameters must be set:\n'
            ) + format_param_list(params)
            raise ConstraintViolated(
                reason, ctx=ctx, constraint=self, params=params)

    def __repr__(self):
        return make_repr(self, self.num_params)


class AcceptBetween(WrapperConstraint):
    def __init__(self, min: int, max: int):  # noqa
        """Satisfied if the number of set parameters is between
        ``min`` and ``max`` (included).

        :param min: must be an integer >= 0
        :param max: must be an integer > min
        """
        check_arg(min >= 0, 'min must be non-negative')
        if max is not None:
            check_arg(min < max, 'must be: min < max.')
        self.min_num_params = min
        self.max_num_params = max
        super().__init__(RequireAtLeast(min) & AcceptAtMost(max), min=min, max=max)

    def help(self, ctx: Context) -> str:
        return f'at least {self.min_num_params} required, ' \
               f'at most {self.max_num_params} accepted'


require_all = _RequireAll()
"""Satisfied if all parameters are set."""

accept_none = AcceptAtMost(0).rephrased(
    help='all forbidden',
    error=f'the following parameters should not be provided:\n'
          f'{ErrorFmt.param_list}'
)
"""Satisfied if none of the parameters is set. Useful only in conditional constraints."""

all_or_none = (require_all | accept_none).rephrased(
    help='provide all or none',
    error=f'the following parameters should be provided together (or none of '
          f'them should be provided):\n'
          f'{ErrorFmt.param_list}',
)
"""Satisfied if either all or none of the parameters are set."""

mutually_exclusive = AcceptAtMost(1).rephrased(
    help='mutually exclusive',
    error=f'the following parameters are mutually exclusive:\n'
          f'{ErrorFmt.param_list}'
)
"""Satisfied if at most one of the parameters is set."""

require_any = RequireAtLeast(1)
"""Alias for ``RequireAtLeast(1)``."""

require_one = RequireExactly(1)
"""Alias for ``RequireExactly(1)``."""



"""
Parameter types and "shortcuts" to create commonly used types.
"""
import pathlib

import click


def path(
    *,
    path_type: type = pathlib.Path,
    exists: bool = False,
    file_okay: bool = True,
    dir_okay: bool = True,
    writable: bool = False,
    readable: bool = True,
    resolve_path: bool = False,
    allow_dash: bool = False,
) -> click.Path:
    """Shortcut for :class:`click.Path` with ``path_type=pathlib.Path``."""
    return click.Path(**locals())


def dir_path(
    *,
    path_type: type = pathlib.Path,
    exists: bool = False,
    writable: bool = False,
    readable: bool = True,
    resolve_path: bool = False,
    allow_dash: bool = False,
) -> click.Path:
    """Shortcut for :class:`click.Path` with
    ``file_okay=False, path_type=pathlib.Path``."""
    return click.Path(**locals(), file_okay=False)


def file_path(
    *,
    path_type: type = pathlib.Path,
    exists: bool = False,
    writable: bool = False,
    readable: bool = True,
    resolve_path: bool = False,
    allow_dash: bool = False,
) -> click.Path:
    """Shortcut for :class:`click.Path` with
    ``dir_okay=False, path_type=pathlib.Path``."""
    return click.Path(**locals(), dir_okay=False)



from typing import Iterable, Sequence, TYPE_CHECKING

import click
from click import Context, Parameter

from .common import join_param_labels

if TYPE_CHECKING:
    from ._core import Constraint


def default_constraint_error(params: Iterable[Parameter], desc: str) -> str:
    return (
        'the following constraint on parameters [%s] was not satisfied: %s'
        % (join_param_labels(params), desc)
    )


class ConstraintViolated(click.UsageError):
    def __init__(
        self, message: str,
        ctx: Context,
        constraint: 'Constraint',
        params: Sequence[click.Parameter]
    ):
        super().__init__(message, ctx=ctx)
        self.ctx = ctx
        self.constraint = constraint
        self.params = params

    @classmethod
    def default(
        cls,
        desc: str,
        ctx: Context,
        constraint: 'Constraint',
        params: Sequence[Parameter],
    ) -> 'ConstraintViolated':
        return ConstraintViolated(
            default_constraint_error(params, desc),
            ctx=ctx, constraint=constraint, params=params,
        )


class UnsatisfiableConstraint(Exception):
    """ Raised if a constraint cannot be satisfied by a group of parameters
    independently from their values at runtime; e.g. mutually_exclusive cannot
    be satisfied if multiple of the parameters are required. """

    def __init__(
        self, constraint: 'Constraint', params: Iterable[Parameter], reason: str
    ):
        self.constraint = constraint
        self.params = params
        self.reason = reason
        param_names = join_param_labels(params)
        message = (f"\nthe constraint {constraint}\n"
                   f"defined on parameters [{param_names}]\n"
                   f"cannot be satisfied because {reason}")
        super().__init__(message)



"""
This modules contains classes to create conditional constraints.
"""
from typing import Optional, Sequence, Union

from click import Context, Parameter

from ._core import Constraint
from .conditions import AllSet, IsSet, Predicate
from .exceptions import ConstraintViolated
from .._util import make_repr


def as_predicate(arg: Union[str, Sequence[str], Predicate]) -> Predicate:
    if isinstance(arg, str):
        return IsSet(arg)
    elif isinstance(arg, Predicate):
        return arg
    elif isinstance(arg, Sequence):
        return AllSet(*arg)
    else:
        raise TypeError('arg should be a string, a list of string or a Predicate')


class If(Constraint):
    def __init__(
        self, condition: Union[str, Sequence[str], Predicate],
        then: Constraint,
        else_: Optional[Constraint] = None
    ):
        """
        Checks one constraint or another depending on the truth value of the condition.

        .. versionadded:: 0.8.0
            you can now pass a sequence of parameter names as condition, which
            corresponds to the predicate ``AllSet(*param_names)``.

        :param condition:
            can be either an instance of ``Predicate`` or (more often) the name of a
            parameter or a list/tuple of parameters that must be all set for the
            condition to be true.
        :param then:
            a constraint checked if the condition is true.
        :param else_:
            an (optional) constraint checked if the condition is false.
        """
        self._condition = as_predicate(condition)
        self._then = then
        self._else = else_

    def help(self, ctx: Context) -> str:
        condition = self._condition.description(ctx)
        then_help = self._then.help(ctx)
        else_help = self._else.help(ctx) if self._else else None
        if not self._else:
            return f'{then_help} if {condition}'
        else:
            return f'{then_help} if {condition}, otherwise {else_help}'

    def check_consistency(self, params: Sequence[Parameter]) -> None:
        self._then.check_consistency(params)
        if self._else:
            self._else.check_consistency(params)

    def check_values(self, params: Sequence[Parameter], ctx: Context):
        condition = self._condition
        condition_is_true = condition(ctx)
        branch = self._then if condition_is_true else self._else
        if branch is None:
            return
        try:
            branch.check_values(params, ctx=ctx)
        except ConstraintViolated as err:
            desc = (condition.description(ctx) if condition_is_true
                    else condition.negated_description(ctx))
            raise ConstraintViolated(
                f"when {desc}, {err}", ctx=ctx, constraint=self, params=params)

    def __repr__(self) -> str:
        if self._else:
            return make_repr(self, self._condition, then=self._then, else_=self._else)
        return make_repr(self, self._condition, then=self._then)



#!/usr/bin/env python
from pathlib import Path
from textwrap import dedent

from setuptools import find_packages, setup


def make_long_description(write_file=False):
    readme = Path('README.rst').read_text(encoding='utf-8')
    # PyPI doesn't support the `raw::` directive. Skip it.
    start = readme.find('.. docs-index-start')
    long_description = readme[start:]
    if write_file:
        Path('PYPI_README.rst').write_text(long_description, encoding='utf-8')
    return long_description


setup(
    name='cloup',
    setup_requires=['setuptools_scm'],
    use_scm_version={
        'write_to': 'cloup/_version.py'
    },
    author='Gianluca Gippetto',
    author_email='gianluca.gippetto@gmail.com',
    description="Adds features to Click: option groups, constraints, subcommand "
                "sections and help themes.",
    long_description_content_type='text/x-rst',
    long_description=make_long_description(),
    url='https://github.com/janLuke/cloup',
    license="BSD 3-Clause",
    keywords=['CLI', 'click', 'argument groups', 'option groups', 'constraints',
              'help colors', 'help themes', 'help styles'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    packages=find_packages(include=['cloup', 'cloup.*']),
    zip_safe=False,
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=[
        'click >=7.1, <9.0',
        'dataclasses; python_version<="3.6"',
        'typing_extensions; python_version<="3.8"',
    ],
)



from itertools import combinations
from typing import Optional
from unittest.mock import Mock

import pytest
from click import Context
from pytest import mark

from cloup.constraints import ConstraintViolated, If
from cloup.constraints.conditions import (
    AllSet, AnySet, Equal, IsSet, Predicate, _And, _Or
)
from tests.constraints.test_constraints import FakeConstraint
from tests.util import make_context, parametrize, mock_repr


class FakePredicate(Predicate):
    def __init__(self, value: bool = True,
                 desc: str = 'description',
                 neg_desc: Optional[str] = None):
        self.value = value
        self._desc = desc
        self._neg_desc = neg_desc

    def description(self, ctx: Context) -> str:
        assert isinstance(ctx, Context)
        return self._desc

    def negated_description(self, ctx: Context) -> str:
        assert isinstance(ctx, Context)
        if self._neg_desc:
            return self._neg_desc
        return super().negated_description(ctx)

    def __call__(self, ctx: Context) -> bool:
        assert isinstance(ctx, Context)
        return self.value


class TestIf:

    def test_help(self, sample_cmd):
        ctx = make_context(sample_cmd, 'arg1 --str-opt=ciao --bool-opt=0')
        condition = FakePredicate(desc='<condition>')
        then_branch = FakeConstraint(help='<a>')
        else_branch = FakeConstraint(help='<b>')

        constr = If(condition, then_branch)
        assert constr.help(ctx) == '<a> if <condition>'

        constr = If(condition, then_branch, else_branch)
        assert constr.help(ctx) == '<a> if <condition>, otherwise <b>'

    @mark.parametrize('condition_is_true', [True, False])
    @mark.parametrize(
        'else_is_provided', [True, False], ids=['else_given', 'else_missing']
    )
    def test_branches_check_methods_are_called_correctly(
        self, sample_cmd, condition_is_true, else_is_provided
    ):
        ctx = make_context(sample_cmd, 'arg1 --str-opt=ciao --bool-opt=0')
        param_names = ['arg1', 'bool_opt']
        params = sample_cmd.get_params_by_name(param_names)

        then_branch = Mock(wraps=FakeConstraint())
        else_branch = Mock(wraps=FakeConstraint()) if else_is_provided else None
        constraint = If(
            condition=FakePredicate(value=condition_is_true),
            then=then_branch,
            else_=else_branch
        )
        constraint.check(param_names, ctx=ctx)

        # check_consistency() is called on both branches whatever the condition value is
        then_branch.check_consistency.assert_called_once_with(params)
        if else_is_provided:
            else_branch.check_consistency.assert_called_once_with(params)

        if condition_is_true:
            then_branch.check_values.assert_called_once_with(params, ctx=ctx)
            if else_is_provided:
                else_branch.check_values.assert_not_called()
        else:
            then_branch.check_values.assert_not_called()
            if else_is_provided:
                else_branch.check_values.assert_called_once_with(params, ctx=ctx)

    def test_error_message(self, sample_cmd):
        ctx = make_context(sample_cmd, 'arg1 --str-opt=ciao --bool-opt=0')
        dummy_params = ['arg1', 'bool_opt']
        then_branch = FakeConstraint(satisfied=False, error='<then_error>')
        else_branch = FakeConstraint(satisfied=False, error='<else_error>')

        true_predicate = FakePredicate(True, desc='<true>')
        with pytest.raises(ConstraintViolated) as info:
            If(true_predicate, then_branch, else_branch).check(dummy_params, ctx=ctx)
        assert info.value.message == 'when <true>, <then_error>'

        false_predicate = FakePredicate(False, desc='<true>', neg_desc='<false>')
        with pytest.raises(ConstraintViolated) as info:
            If(false_predicate, then_branch, else_branch).check(dummy_params, ctx=ctx)
        assert info.value.message == 'when <false>, <else_error>'

    def test_init_with_string_as_condition(self):
        constr = If('name', FakeConstraint())
        assert isinstance(constr._condition, IsSet)
        assert constr._condition.param_name == 'name'

    def test_init_with_sequence_of_strings_as_condition(self):
        param_names = ('opt1', 'opt2')
        constr = If(param_names, FakeConstraint())
        assert isinstance(constr._condition, AllSet)
        assert constr._condition.param_names == param_names

    def test_repr(self):
        predicate = mock_repr('<Predicate>', spec=FakePredicate, wraps=FakePredicate())
        then_branch = mock_repr('<ThenBranch>', wraps=FakeConstraint())
        else_branch = mock_repr('<ElseBranch>', wraps=FakeConstraint())

        constr = If(predicate, then_branch)
        assert repr(constr) == 'If(<Predicate>, then=<ThenBranch>)'
        constr = If(predicate, then_branch, else_branch)
        assert repr(constr) == 'If(<Predicate>, then=<ThenBranch>, else_=<ElseBranch>)'


class TestNot:
    def test_descriptions(self, dummy_ctx):
        neg = ~FakePredicate(desc='<desc>')
        assert neg.desc(dummy_ctx) == 'NOT(<desc>)'
        assert neg.neg_desc(dummy_ctx) == '<desc>'

        neg = ~FakePredicate(desc='<desc>', neg_desc='<neg_desc>')
        assert neg.desc(dummy_ctx) == '<neg_desc>'
        assert neg.neg_desc(dummy_ctx) == '<desc>'

    def test_double_negation_returns_original_predicate(self):
        fake = FakePredicate()
        assert ~~fake is fake


class TestAnd:
    @mark.parametrize('b_value', [False, True])
    @mark.parametrize('a_value', [False, True])
    def test_evaluation(self, a_value, b_value, dummy_ctx):
        a = FakePredicate(value=a_value)
        b = FakePredicate(value=b_value)
        c = a & b
        assert c(dummy_ctx) == (a_value and b_value)

    def test_operand_merging(self):
        a, b, c, d = (FakePredicate() for _ in range(4))
        res = (a & b) & c
        assert res.predicates == (a, b, c)
        res = (a & b) & (c & d)
        assert res.predicates == (a, b, c, d)
        res = (a & b) & (c | d)
        assert len(res.predicates) == 3

    def test_descriptions(self, dummy_ctx):
        a, b, c = (FakePredicate(desc=name) for name in 'ABC')
        assert (a & b & c).description(dummy_ctx) == 'A and B and C'
        assert (a & b & c).neg_desc(dummy_ctx) == 'NOT(A) or NOT(B) or NOT(C)'


class TestOr:
    @mark.parametrize('b_value', [False, True])
    @mark.parametrize('a_value', [False, True])
    def test_evaluation(self, a_value, b_value, dummy_ctx):
        a = FakePredicate(value=a_value)
        b = FakePredicate(value=b_value)
        c = a | b
        assert c(dummy_ctx) == (a_value or b_value)

    def test_operands_merging(self):
        a, b, c, d = (FakePredicate() for _ in range(4))
        res = (a | b) | c
        assert res.predicates == (a, b, c)
        res = (a | b) | (c | d)
        assert res.predicates == (a, b, c, d)
        res = (a | b) | (c & d)
        assert len(res.predicates) == 3

    def test_descriptions(self, dummy_ctx):
        a, b, c = (FakePredicate(desc=desc) for desc in 'ABC')
        assert (a | b | c).description(dummy_ctx) == 'A or B or C'
        assert (a | b | c).neg_desc(dummy_ctx) == 'NOT(A) and NOT(B) and NOT(C)'


def test_description_with_mixed_operators(dummy_ctx):
    ctx = dummy_ctx
    a, b, c, d = (FakePredicate(desc=desc) for desc in 'ABCD')
    assert (a & b & c).desc(ctx) == 'A and B and C'
    assert (a | b | c).desc(ctx) == 'A or B or C'
    assert (a | b & c).desc(ctx) == 'A or (B and C)'
    assert (a & b | c).desc(ctx) == '(A and B) or C'
    assert ((a | b) & (c | d)).desc(ctx) == '(A or B) and (C or D)'


class TestIsSet:
    SHELL_INPUT = 'arg1 --bool-opt=0 --flag2 --mul1 1 --mul1 2'

    @parametrize(
        # These cases are relative to [sample_cmd] with [SHELL_INPUT] as input
        # "provided" means provided by the user in the command line
        ['param', 'is_set'],
        pytest.param('arg1',     True,  id='provided argument'),
        pytest.param('arg2',     False, id='unset argument'),
        pytest.param('bool_opt', True,  id='provided option'),
        pytest.param('str_opt',  False, id='unprovided option without default'),
        pytest.param('def1',     True,  id='unprovided option with default'),
        pytest.param('flag',     False, id='unset boolean flag'),
        pytest.param('flag2',    True,  id='provided boolean flag'),
        pytest.param('mul1',     True,  id='provided multi-option'),
        pytest.param('mul2',     False, id='unprovided multi-option'),
        pytest.param('tuple',    False, id='unprovided option with nargs>1'),
    )
    def test_evaluation(self, sample_cmd, param, is_set):
        ctx = make_context(sample_cmd, self.SHELL_INPUT)
        assert IsSet(param)(ctx) == is_set
        assert ~IsSet(param)(ctx) != is_set

    def test_descriptions(self, sample_cmd):
        ctx = make_context(sample_cmd, '')
        assert IsSet('arg1').desc(ctx) == 'ARG1 is set'
        assert IsSet('str_opt').desc(ctx) == '--str-opt is set'
        assert IsSet('arg1').neg_desc(ctx) == 'ARG1 is not set'
        assert IsSet('str_opt').neg_desc(ctx) == '--str-opt is not set'

    def test_and(self):
        a, b, c = IsSet('opt1'), IsSet('opt2'), Equal('opt2', 'value')
        assert a & b == AllSet('opt1', 'opt2')
        assert a & c == _And(a, c)

    def test_or(self):
        a, b, c = IsSet('opt1'), IsSet('opt2'), Equal('opt2', 'value')
        assert a | b == AnySet('opt1', 'opt2')
        assert a | c == _Or(a, c)

    def test_eq(self):
        assert IsSet('a') == IsSet('a')
        assert IsSet('a') != IsSet('b')
        assert IsSet('a') != Equal('a', 'ciao')


class TestEqual:
    def test_evaluation(self, sample_cmd):
        ctx = make_context(sample_cmd, 'xxx --bool-opt=0 --flag --tuple 1 2')
        for name, value in ctx.params.items():
            assert Equal(name, value)(ctx)
            assert not Equal(name, 'blah')(ctx)
            assert not (~Equal(name, value))(ctx)
            assert ~Equal(name, 'blah')(ctx)

    def test_descriptions(self, sample_cmd):
        ctx = make_context(sample_cmd, '')
        p = Equal('arg1', 'value')
        assert p.desc(ctx) == 'ARG1="value"'
        assert p.neg_desc(ctx) == 'ARG1!="value"'

    def test_eq(self):
        assert Equal('name', 'value') == Equal('name', 'value')
        assert Equal('name', 'foo') != Equal('name', 'bar')
        assert Equal('foo', 'value') != Equal('bar', 'value')


class TestAllSet:
    def test_evaluation_of_true_predicates(self, sample_cmd):
        ctx = make_context(sample_cmd, 'xxx --bool-opt=0 --flag --tuple 1 2')
        set_params = ['arg1', 'bool_opt', 'flag', 'tuple']
        for n in range(1, len(set_params)):
            for param_group in combinations(set_params, n):
                true = AllSet(*param_group)
                assert true(ctx), param_group
                assert not (~true)(ctx), param_group

    def test_evaluation_of_false_predicates(self, sample_cmd):
        ctx = make_context(sample_cmd, 'xxx --bool-opt=0 --flag --tuple 1 2')
        param_groups = [('arg1', 'arg2'), ('arg1', 'int_opt'), ('arg2', 'str_opt')]
        for param_group in param_groups:
            false = AllSet(*param_group)
            assert not false(ctx), param_group
            assert (~false)(ctx), param_group

    def test_descriptions(self, sample_cmd):
        ctx = make_context(sample_cmd, '')
        assert AllSet('arg1').desc(ctx) == 'ARG1 is set'
        assert AllSet('arg1').neg_desc(ctx) == 'ARG1 is not set'

        assert AllSet('arg1', 'flag').desc(ctx) \
               == 'ARG1 and --flag are both set'
        assert AllSet('arg1', 'flag').neg_desc(ctx) \
               == 'ARG1 and --flag are not both set'

        assert AllSet('arg1', 'flag', 'int_opt').desc(ctx) \
               == 'ARG1, --flag and --int-opt are all set'
        assert AllSet('arg1', 'flag', 'int_opt').neg_desc(ctx) \
               == 'ARG1, --flag and --int-opt are not all set'

    def test_and(self):
        allset1 = AllSet('a', 'b')
        allset2 = AllSet('c', 'd')
        anyset = AnySet('foo', 'bar')
        assert allset1 & allset2 == AllSet('a', 'b', 'c', 'd')
        assert allset1 & anyset == _And(allset1, anyset)

    def test_eq(self):
        assert AllSet('a', 'b') == AllSet('a', 'b')
        assert AllSet('a') != AllSet('a', 'b')
        assert AllSet('a', 'b') != AnySet('a', 'b')


class TestAnySet:
    def test_evaluation(self, sample_cmd):
        ctx = make_context(sample_cmd, 'xxx --bool-opt=0 --flag --tuple 1 2')
        set_params = ['arg1', 'bool_opt', 'flag', 'tuple']
        unset_params = ['arg2', 'int_opt', 'str_opt']
        for set_param in set_params:
            true = AnySet(*unset_params, set_param)
            assert true(ctx)
            assert not (~true)(ctx)
        false = AnySet(*unset_params)
        assert not false(ctx)
        assert (~false)(ctx)

    def test_descriptions(self, sample_cmd):
        ctx = make_context(sample_cmd, '')
        assert AnySet('arg1').desc(ctx) == 'ARG1 is set'
        assert AnySet('arg1').neg_desc(ctx) == 'ARG1 is not set'

        assert AnySet('arg1', 'flag').desc(ctx) \
               == 'either ARG1 or --flag is set'
        assert AnySet('arg1', 'flag').neg_desc(ctx) \
               == 'neither ARG1 nor --flag is set'

        assert AnySet('arg1', 'flag', 'int_opt').desc(ctx) \
               == 'any of ARG1, --flag and --int-opt is set'
        assert AnySet('arg1', 'flag', 'int_opt').neg_desc(ctx) \
               == 'none of ARG1, --flag and --int-opt is set'

    def test_or(self):
        anyset1 = AnySet('a', 'b')
        anyset2 = AnySet('c', 'd')
        allset = AllSet('foo', 'bar')
        assert anyset1 | anyset2 == AnySet('a', 'b', 'c', 'd')
        assert anyset1 | allset == _Or(anyset1, allset)

    def test_eq(self):
        assert AnySet('a', 'b') == AnySet('a', 'b')
        assert AnySet('a') != AnySet('a', 'b')
        assert AnySet('a', 'b') != AllSet('a', 'b')



"""Tests for the "subcommand sections" feature/module."""
import click
import pytest
from click import pass_context

import cloup
from cloup import Section
from cloup._util import pick_non_missing, reindent
from cloup.typing import MISSING
from tests.util import new_dummy_func, pick_first_bool


@pytest.mark.parametrize(
    'align_sections', [True, False], ids=['aligned', 'non-aligned']
)
def test_subcommand_sections_are_correctly_rendered_in_help(
    runner, align_sections, get_example_group
):
    grp = get_example_group(align_sections)
    result = runner.invoke(grp, args=('--help',))
    if result.exception:
        raise result.exception
    assert result.exit_code == 0
    assert result.output.strip() == grp.expected_help


@pytest.mark.parametrize(
    'subcommand_cls', [click.Command, cloup.Command, click.Group, cloup.Group],
    ids=['click_Command', 'cloup_Command', 'click_Group', 'cloup_Group'],
)
@pytest.mark.parametrize(
    'assign_to_section', [True, False],
    ids=['with_section', 'without_section'],
)
def test_Group_subcommand_decorator(subcommand_cls, assign_to_section):
    grp = cloup.Group('name')
    # Use @grp.group if subcommand_class is a Group, else @grp.Command
    method_name = ('group' if issubclass(subcommand_cls, cloup.Group)
                   else 'command')
    decorator = getattr(grp, method_name)
    # Add a subcommand to the Group using the decorator
    subcommand_name = 'cmd'
    section_arg = Section('title') if assign_to_section else None
    subcommand = decorator(
        subcommand_name,
        section=section_arg,
        cls=subcommand_cls,
        help='Help'
    )(new_dummy_func())
    assert subcommand.__class__ is subcommand_cls
    assert subcommand.help == 'Help'
    assert grp.commands[subcommand_name] is subcommand
    if assign_to_section:
        section = grp._user_sections[0]
        assert section is section_arg
        assert section.commands[subcommand_name] is subcommand
    else:
        assert grp._default_section.commands[subcommand_name] is subcommand


@pytest.mark.parametrize(
    'cmd_value', [MISSING, None, True, False],
    ids=lambda val: f'cmd_{val}'
)
@pytest.mark.parametrize(
    'ctx_value', [MISSING, None, True, False],
    ids=lambda val: f'ctx_{val}'
)
def test_align_sections_context_setting(runner, ctx_value, cmd_value):
    should_align = pick_first_bool([cmd_value, ctx_value], default=True)
    cxt_settings = pick_non_missing(dict(
        align_sections=ctx_value,
        terminal_width=80,
    ))
    cmd_kwargs = pick_non_missing(dict(
        align_sections=cmd_value,
        context_settings=cxt_settings
    ))

    @cloup.group(**cmd_kwargs)
    @pass_context
    def cmd(ctx, one, much_longer_opt):
        assert cmd.must_align_sections(ctx) == should_align

    cmd.section(
        "First section",
        cloup.command('cmd', help='First command help')(new_dummy_func()),
    )

    cmd.section(
        "Second section",
        cloup.command('longer-cmd', help='Second command help')(new_dummy_func()),
    )

    result = runner.invoke(cmd, args=('--help',))
    start = result.output.find('First section')
    if should_align:
        expected = """
            First section:
              cmd         First command help

            Second section:
              longer-cmd  Second command help"""
    else:
        expected = """
            First section:
              cmd  First command help

            Second section:
              longer-cmd  Second command help"""

    expected = reindent(expected)
    end = start + len(expected)
    assert result.output[start:end] == expected


def test_override_format_subcommand_name(runner):
    class MyGroup(cloup.Group):
        def format_subcommand_name(self, ctx, name, cmd) -> str:
            return '*special*' if name == 'special' else name

    main = MyGroup(name='main')
    main.section(
        'Commands',
        cloup.Command(name='special', help='A special command.'),
        cloup.Command(name='ordinary', help='An ordinary command.')
    )

    res = runner.invoke(main, ['--help'])
    expected_help = reindent("""
        Usage: main [OPTIONS] COMMAND [ARGS]...

        Options:
          --help  Show this message and exit.

        Commands:
          *special*  A special command.
          ordinary   An ordinary command.
    """)
    assert res.output == expected_help


def test_section_error_if_first_arg_is_not_a_string():
    with pytest.raises(TypeError, match="the first argument must be a string"):
        Section([cloup.Command('cmd')])
    grp = cloup.Group()
    with pytest.raises(TypeError, match="the first argument must be a string"):
        grp.section([cloup.Command('cmd')])



"""
Tip: in your editor, set a ruler at 80 characters.
"""
import inspect
from textwrap import dedent
from typing import Optional

import click
import pytest

from cloup import HelpFormatter
from cloup.typing import Possibly
from cloup.formatting import HelpSection, unstyled_len
from cloup.formatting.sep import (
    Hline, RowSepIf, RowSepPolicy, multiline_rows_are_at_least
)
from cloup.styling import HelpTheme, Style
from tests.util import parametrize

LOREM = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor."
ROWS = [
    ('-l, --long-option-name TEXT', LOREM),
    ('--another-option INT', LOREM),
    ('--short', LOREM),
]


def test_write_dl_with_col1_max_width_equal_to_longest_col1_value():
    formatter = HelpFormatter(width=80, col1_max_width=len(ROWS[0][0]))
    formatter.current_indent = 4
    expected = """
    -l, --long-option-name TEXT  Lorem ipsum dolor sit amet, consectetur
                                 adipiscing elit, sed do eiusmod tempor.
    --another-option INT         Lorem ipsum dolor sit amet, consectetur
                                 adipiscing elit, sed do eiusmod tempor.
    --short                      Lorem ipsum dolor sit amet, consectetur
                                 adipiscing elit, sed do eiusmod tempor.
    """[1:-4]
    formatter.write_dl(ROWS)
    assert formatter.getvalue() == expected


def test_formatter_excludes_rows_exceeding_col1_max_width_from_col1_width_computation():
    formatter = HelpFormatter(width=80, col1_max_width=len('--short'))
    formatter.current_indent = 4
    expected = """
    -l, --long-option-name TEXT
             Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do
             eiusmod tempor.
    --another-option INT
             Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do
             eiusmod tempor.
    --short  Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do
             eiusmod tempor.
    """[1:-4]
    formatter.write_dl(ROWS)
    assert formatter.getvalue() == expected


def test_col2_min_width():
    formatter = HelpFormatter(width=80)
    formatter.current_indent = 4
    formatter.col2_min_width = (
        formatter.available_width - len(ROWS[0][0]) - formatter.col_spacing)
    expected = """
    -l, --long-option-name TEXT  Lorem ipsum dolor sit amet, consectetur
                                 adipiscing elit, sed do eiusmod tempor.
    --another-option INT         Lorem ipsum dolor sit amet, consectetur
                                 adipiscing elit, sed do eiusmod tempor.
    --short                      Lorem ipsum dolor sit amet, consectetur
                                 adipiscing elit, sed do eiusmod tempor.
    """[1:-4]
    formatter.write_dl(ROWS)
    assert formatter.getvalue() == expected

    formatter.buffer = []
    formatter.col2_min_width += 1
    expected = """
    -l, --long-option-name TEXT
       Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod
       tempor.

    --another-option INT
       Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod
       tempor.

    --short
       Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod
       tempor.
    """[1:-4]
    formatter.write_dl(ROWS)
    assert formatter.getvalue() == expected


def test_value_error_if_row_sep_string_ends_with_newline():
    with pytest.raises(ValueError, match=r"row_sep must not end with '\\n'"):
        HelpFormatter(row_sep='\n')


@parametrize(
    'row_sep',
    pytest.param('-' * (80 - 4), id='string'),
    pytest.param(Hline.dashed, id='SepGenerator'),
)
def test_fixed_row_sep(row_sep):
    formatter = HelpFormatter(row_sep=row_sep, width=80)
    formatter.current_indent = 4
    expected = """
    -l, --long-option-name TEXT  Lorem ipsum dolor sit amet, consectetur
                                 adipiscing elit, sed do eiusmod tempor.
    ----------------------------------------------------------------------------
    --another-option INT         Lorem ipsum dolor sit amet, consectetur
                                 adipiscing elit, sed do eiusmod tempor.
    ----------------------------------------------------------------------------
    --short                      Lorem ipsum dolor sit amet, consectetur
                                 adipiscing elit, sed do eiusmod tempor.
    """[1:-4]
    formatter.write_dl(ROWS)
    actual = formatter.getvalue()
    assert actual == expected


@parametrize(
    ['policy', 'expected_sep'],  # unless None, expected_sep should end with \n
    # Three of the test rows are "multi-line"
    pytest.param(
        RowSepIf(multiline_rows_are_at_least(3)),
        '',
        id='empty_line'),
    pytest.param(
        RowSepIf(multiline_rows_are_at_least(3), sep=Hline.dashed),
        Hline.dashed(80),
        id='dashed_line'),
    pytest.param(
        RowSepIf(multiline_rows_are_at_least(4)),
        None,
        id='no_sep'),
)
def test_conditional_row_sep(policy: RowSepPolicy, expected_sep: Optional[str]):
    formatter = HelpFormatter(
        width=80, col1_max_width=30, col2_min_width=0, col_spacing=3,
        row_sep=policy,
    )
    rows = [
        ('Longest than 30 characters for sure', 'Short help'),
        ('Longest than 30 characters for sure', 'Another short help'),
        ('Below 30 characters', LOREM),
        ('Short help', 'Short help'),
    ]
    actual_expected_sep = '' if expected_sep is None else (expected_sep + '\n')
    expected = """
Longest than 30 characters for sure
                      Short help
---
Longest than 30 characters for sure
                      Another short help
---
Below 30 characters   Lorem ipsum dolor sit amet, consectetur adipiscing elit,
                      sed do eiusmod tempor.
---
Short help            Short help
    """[1:-4].replace('---\n', actual_expected_sep)

    formatter.write_dl(rows)
    actual = formatter.getvalue()
    assert actual == expected


def test_formatter_settings_creation():
    assert HelpFormatter.settings() == {}
    assert HelpFormatter.settings(
        indent_increment=4, col_spacing=3
    ) == dict(indent_increment=4, col_spacing=3)


def test_settings_signature_matches_HelpFormatter():
    cls_params = dict(inspect.signature(HelpFormatter).parameters)
    settings = dict(inspect.signature(HelpFormatter.settings).parameters)
    assert set(cls_params) == set(settings)
    # Check type annotations
    for name, param in cls_params.items():
        assert settings[name].annotation == Possibly[cls_params[name].annotation], name


def test_write_heading():
    formatter = HelpFormatter(theme=HelpTheme(
        heading=Style(fg='blue', bold=True),
    ))
    formatter.write_heading('Title')
    assert formatter.getvalue() == click.style('Title:', fg='blue', bold=True) + '\n'


def test_write_text_with_styles():
    formatter = HelpFormatter(width=80)
    formatter.current_indent = 4
    WRAPPED = dedent("""
        Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod
        tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam,
        quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo
        consequat."""[1:])
    INPUT_TEXT = ' '.join(WRAPPED.split())
    indentation = ' ' * formatter.current_indent
    style = Style(fg='yellow')
    EXPECTED = '\n'.join(
        indentation + style(line) for line in WRAPPED.splitlines()
    )
    formatter.write_text(INPUT_TEXT, style=style)
    actual = formatter.getvalue().rstrip()
    for line in actual.splitlines():
        assert unstyled_len(line) <= formatter.width
    assert actual == EXPECTED


def test_write_section_print_long_constraint_on_a_new_line():
    formatter = HelpFormatter(width=72, indent_increment=4)
    section = HelpSection(
        'The heading',
        [('term', 'This is the definition.')],
        help='This is the help.',
        constraint="""
            This is a long constraint description that doesn't fit the line xx.
        """.strip()
    )
    expected = dedent("""
        The heading:
            [This is a long constraint description that doesn't fit the line
            xx.]
            This is the help.
            term  This is the definition.
        """)
    formatter.write_section(section)
    actual = formatter.getvalue()
    assert actual == expected



"""
This modules contains predicates with an associated description that you can use
as conditions of conditional constraints (see :class:`cloup.constraints.If`).

Predicates should be treated as immutable objects, even though immutability
is not (at the moment) enforced.
"""
import abc
from typing import Any, Generic, TypeVar

from click import Context

from ._support import ensure_constraints_support
from .common import (
    get_param_labels,
    get_param_name,
    join_with_and,
    param_label_by_name,
    param_value_by_name,
    param_value_is_set,
)
from .._util import make_repr

P = TypeVar('P', bound='Predicate')


class Predicate(abc.ABC):
    """
    A ``Callable`` that takes a ``Context`` and returns a boolean, with an
    associated description. Meant to be used as condition in a conditional
    constraint (see :class:`~cloup.constraints.If`).
    """

    @abc.abstractmethod
    def description(self, ctx: Context) -> str:
        """Succint description of the predicate (alias: `desc`)."""

    def negated_description(self, ctx: Context) -> str:
        """Succint description of the negation of this predicate (alias: `neg_desc`)."""
        return 'NOT(%s)' % self.description(ctx)

    def desc(self, ctx: Context) -> str:
        """Short alias for :meth:`description`."""
        return self.description(ctx)

    def neg_desc(self, ctx: Context) -> str:
        """Short alias for :meth:`negated_description`."""
        return self.negated_description(ctx)

    def negated(self):
        return ~self

    @abc.abstractmethod
    def __call__(self, ctx: Context) -> bool:
        """Evaluate the predicate on the given context."""

    def __invert__(self) -> 'Predicate':
        return Not(self)

    def __or__(self, other: 'Predicate') -> '_Or':
        return _Or(self, other)

    def __and__(self, other: 'Predicate') -> '_And':
        return _And(self, other)

    def __repr__(self):
        return make_repr(self, *self._public_fields().values())

    def _public_fields(self):
        return {k: v for k, v in vars(self).items() if not k.startswith('_')}

    def __eq__(self, other):
        return isinstance(other, self.__class__) and (
            self._public_fields() == other._public_fields()
        )


class Not(Predicate, Generic[P]):
    """Logical NOT of a predicate."""
    def __init__(self, predicate: P):
        self.predicate = predicate

    def description(self, ctx: Context) -> str:
        return self.predicate.negated_description(ctx)

    def negated_description(self, ctx: Context) -> str:
        return self.predicate.description(ctx)

    def __call__(self, ctx: Context) -> bool:
        return not self.predicate(ctx)

    def __invert__(self) -> P:
        return self.predicate

    def __repr__(self) -> str:
        return 'Not(%r)' % self.predicate


class _Operator(Predicate, metaclass=abc.ABCMeta):
    """Operator between two or more predicates."""
    DESC_SEP: str

    def __init__(self, *predicates: Predicate):
        if len(predicates) < 2:
            raise ValueError('provide at least 2 predicates')
        self.predicates = predicates

    def description(self, ctx: Context) -> str:
        return self.DESC_SEP.join(
            '(%s)' % p.description(ctx) if isinstance(p, _Operator)
            else p.description(ctx)
            for p in self.predicates
        )

    def __repr__(self):
        return make_repr(self, *self.predicates)


class _And(_Operator):
    """Logical AND of two or more predicates."""
    DESC_SEP = ' and '

    def negated_description(self, ctx: Context) -> str:
        return ' or '.join(
            '(%s)' % p.neg_desc(ctx) if isinstance(p, _Operator) else p.neg_desc(ctx)
            for p in self.predicates
        )

    def __call__(self, ctx: Context) -> bool:
        return all(p(ctx) for p in self.predicates)

    def __and__(self, other: 'Predicate') -> '_And':
        if isinstance(other, _And):
            return _And(*self.predicates, *other.predicates)
        return _And(*self.predicates, other)


class _Or(_Operator):
    """Logical OR of two or more predicates."""
    DESC_SEP = ' or '

    def negated_description(self, ctx: Context) -> str:
        return ' and '.join(
            '(%s)' % p.neg_desc(ctx) if isinstance(p, _Operator) else p.neg_desc(ctx)
            for p in self.predicates
        )

    def __call__(self, ctx: Context) -> bool:
        return any(p(ctx) for p in self.predicates)

    def __or__(self, other: 'Predicate') -> '_Or':
        if isinstance(other, _Or):
            return _Or(*self.predicates, *other.predicates)
        return _Or(*self.predicates, other)


class IsSet(Predicate):
    """True if the parameter is set."""
    def __init__(self, param_name: str):
        self.param_name = param_name

    def description(self, ctx: Context) -> str:
        return '%s is set' % param_label_by_name(ctx, self.param_name)

    def negated_description(self, ctx: Context) -> str:
        return '%s is not set' % param_label_by_name(ctx, self.param_name)

    def __call__(self, ctx: Context) -> bool:
        command = ensure_constraints_support(ctx.command)
        param = command.get_param_by_name(self.param_name)
        value = param_value_by_name(ctx, self.param_name)
        return param_value_is_set(param, value)

    def __and__(self, other: Predicate):
        if isinstance(other, IsSet):
            return AllSet(self.param_name, other.param_name)
        return super().__and__(other)

    def __or__(self, other: Predicate):
        if isinstance(other, IsSet):
            return AnySet(self.param_name, other.param_name)
        return super().__or__(other)


class AllSet(Predicate):
    """True if all listed parameters are set.

    .. versionadded:: 0.8.0
    """
    def __init__(self, *param_names: str):
        if not param_names:
            raise ValueError('you must provide at least one param name')
        self.param_names = param_names

    def negated_description(self, ctx: Context) -> str:
        labels = get_param_labels(ctx, self.param_names)
        if len(labels) == 1:
            return f'{labels[0]} is not set'
        pronoun = 'both' if len(labels) == 2 else 'all'
        return f'{join_with_and(labels)} are not {pronoun} set'

    def description(self, ctx: Context) -> str:
        labels = get_param_labels(ctx, self.param_names)
        if len(labels) == 1:
            return f'{labels[0]} is set'
        pronoun = 'both' if len(labels) == 2 else 'all'
        return f'{join_with_and(labels)} are {pronoun} set'

    def __call__(self, ctx: Context) -> bool:
        command = ensure_constraints_support(ctx.command)
        params = command.get_params_by_name(self.param_names)
        return all(param_value_is_set(param, ctx.params[get_param_name(param)])
                   for param in params)

    def __and__(self, other: Predicate):
        if isinstance(other, AllSet):
            return AllSet(*self.param_names, *other.param_names)
        return super().__and__(other)


class AnySet(Predicate):
    """True if any of the listed parameters is set.

    .. versionadded:: 0.8.0
    """
    def __init__(self, *param_names: str):
        if not param_names:
            raise ValueError('you must provide at least one param name')
        self.param_names = param_names

    def negated_description(self, ctx: Context) -> str:
        labels = get_param_labels(ctx, self.param_names)
        if len(labels) == 1:
            return f'{labels[0]} is not set'
        if len(labels) == 2:
            return 'neither {} nor {} is set'.format(*labels)
        return f'none of {join_with_and(labels)} is set'

    def description(self, ctx: Context) -> str:
        labels = get_param_labels(ctx, self.param_names)
        if len(labels) == 1:
            return f'{labels[0]} is set'
        if len(labels) == 2:
            return 'either {} or {} is set'.format(*labels)
        return f'any of {join_with_and(labels)} is set'

    def __call__(self, ctx: Context) -> bool:
        command = ensure_constraints_support(ctx.command)
        params = command.get_params_by_name(self.param_names)
        return any(param_value_is_set(param, ctx.params[get_param_name(param)])
                   for param in params)

    def __or__(self, other: Predicate):
        if isinstance(other, AnySet):
            return AnySet(*self.param_names, *other.param_names)
        return super().__or__(other)


class Equal(Predicate):
    """True if the parameter value equals ``value``."""

    def __init__(self, param_name: str, value: Any):
        self.param_name = param_name
        self.value = value

    def description(self, ctx: Context) -> str:
        param_label = param_label_by_name(ctx, self.param_name)
        return f'{param_label}="{self.value}"'

    def negated_description(self, ctx: Context) -> str:
        param_label = param_label_by_name(ctx, self.param_name)
        return f'{param_label}!="{self.value}"'

    def __call__(self, ctx: Context) -> bool:
        return param_value_by_name(ctx, self.param_name) == self.value



"""
Useful functions used to implement constraints and predicates.
"""
from typing import Any, Dict, Iterable, List, Sequence

from click import Argument, Context, Option, Parameter


def param_value_is_set(param: Parameter, value: Any) -> bool:
    """Defines what it means for a parameter of a specific kind to be "set".

    All cases are obvious besides that of boolean options:
    - (common rule) if the value is ``None``, the parameter is unset;
    - a parameter that takes multiple values is set if at least one argument is provided;
    - a boolean **flag** is set only if True;
    - a boolean option is set if not None, even if it's False.
    """
    if value is None:
        return False
    # checking for param.is_flag is redundant but necessary to work around
    # Click 8.0.1 issue: https://github.com/pallets/click/issues/1925
    elif isinstance(param, Option) and param.is_flag and param.is_bool_flag:
        return value
    elif param.nargs != 1 or param.multiple:
        return len(value) > 0
    return True


def get_param_name(param: Parameter) -> str:
    """Returns the name of a parameter casted to ``str``.
    Use this function to avoid typing errors in places where you expect a parameter
    having a name.
    """
    if param.name is None:
        raise TypeError(
            'param.name is required to be a string in this context.\n'
            'Hint: param.name is None only when parameter.expose_value is False, '
            'so you are probably using such option incorrectly.'
        )
    return param.name


def get_params_whose_value_is_set(
    params: Iterable[Parameter], values: Dict[str, Any]
) -> List[Parameter]:
    """Filters ``params`` returning only the parameters that have a value.
    Boolean flags are considered "set" if their value is ``True``."""
    return [p for p in params
            if param_value_is_set(p, values[get_param_name(p)])]


def get_required_params(params: Iterable[Parameter]) -> List[Parameter]:
    return [p for p in params if p.required]


def get_param_label(param: Parameter) -> str:
    if param.param_type_name == 'argument':
        return param.human_readable_name
    return sorted(param.opts, key=len)[-1]


def join_param_labels(params: Iterable[Parameter], sep: str = ', ') -> str:
    return sep.join(get_param_label(p) for p in params)


def join_with_and(strings: Sequence[str], sep: str = ', ') -> str:
    if not strings:
        return ''
    if len(strings) == 1:
        return strings[0]
    return sep.join(strings[:-1]) + ' and ' + strings[-1]


def format_param(param: Parameter) -> str:
    if isinstance(param, Argument):
        return param.human_readable_name
    # return ' | '.join(param.opts)
    if len(param.opts) == 1:
        return param.opts[0]
    short, long = sorted(param.opts, key=len)
    return f'{long} ({short})'


def format_param_list(param_list: Iterable[Parameter], indent=2) -> str:
    lines = map(format_param, param_list)
    indentation = ' ' * indent
    return ''.join(indentation + line + '\n'
                   for line in lines)


def param_label_by_name(ctx, name: str) -> str:
    return get_param_label(ctx.command.get_param_by_name(name))


def get_param_labels(ctx, param_names: Iterable[str]) -> List[str]:
    params = ctx.command.get_params_by_name(param_names)
    return [get_param_label(param) for param in params]


def param_value_by_name(ctx: Context, name: str) -> Any:
    try:
        return ctx.params[name]
    except KeyError:
        raise KeyError(f'"{name}" is not the name of a CLI parameter')



"""
This module contains components that specifically address the styling and theming
of the ``--help`` output.
"""
import dataclasses as dc
from typing import Callable, NamedTuple, Optional

import click

from cloup._util import FrozenSpace, click_version_tuple, delete_keys, identity

IStyle = Callable[[str], str]
"""A callable that takes a string and returns a styled version of it."""


# noinspection PyUnresolvedReferences
class HelpTheme(NamedTuple):
    """A collection of styles for several elements of the help page.

    A "style" is just a function or a callable that takes a string and returns
    a styled version of it. This means you can use your favorite styling/color
    library (like rich, colorful etc). Nonetheless, given that Click has some
    basic styling functionality built-in, Cloup provides the :class:`Style`
    class, which is a wrapper of the ``click.style`` function.

    :param invoked_command:
        Style of the invoked command name (in Usage).
    :param command_help:
        Style of the invoked command description (below Usage).
    :param heading:
        Style of help section headings.
    :param constraint:
        Style of an option group constraint description.
    :param section_help:
        Style of the help text of a section (the optional paragraph below the heading).
    :param col1:
        Style of the first column of a definition list (options and command names).
    :param col2:
        Style of the second column of a definition list (help text).
    :param epilog:
        Style of the epilog.
    """

    invoked_command: IStyle = identity
    """Style of the invoked command name (in Usage)."""

    command_help: IStyle = identity
    """Style of the invoked command description (below Usage)."""

    heading: IStyle = identity
    """Style of help section headings."""

    constraint: IStyle = identity
    """Style of an option group constraint description."""

    section_help: IStyle = identity
    """Style of the help text of a section (the optional paragraph below the heading)."""

    col1: IStyle = identity
    """Style of the first column of a definition list (options and command names)."""

    col2: IStyle = identity
    """Style of the second column of a definition list (help text)."""

    epilog: IStyle = identity
    """Style of the epilog."""

    def with_(
        self, invoked_command: Optional[IStyle] = None,
        command_help: Optional[IStyle] = None,
        heading: Optional[IStyle] = None,
        constraint: Optional[IStyle] = None,
        section_help: Optional[IStyle] = None,
        col1: Optional[IStyle] = None,
        col2: Optional[IStyle] = None,
        epilog: Optional[IStyle] = None,
    ) -> 'HelpTheme':
        kwargs = {key: val for key, val in locals().items() if val is not None}
        kwargs.pop('self')
        if kwargs:
            return self._replace(**kwargs)
        return self

    @staticmethod
    def dark():
        """A theme assuming a dark terminal background color."""
        return HelpTheme(
            invoked_command=Style(fg='bright_yellow'),
            heading=Style(fg='bright_white', bold=True),
            constraint=Style(fg='magenta'),
            col1=Style(fg='bright_yellow'),
        )

    @staticmethod
    def light():
        """A theme assuming a light terminal background color."""
        return HelpTheme(
            invoked_command=Style(fg='yellow'),
            heading=Style(fg='bright_blue'),
            constraint=Style(fg='red'),
            col1=Style(fg='yellow'),
        )


@dc.dataclass(frozen=True)
class Style:
    """Wraps :func:`click.style` for a better integration with :class:`HelpTheme`.

    Available colors are defined as static constants in :class:`Color`.

    Arguments are set to ``None`` by default. Passing ``False`` to boolean args
    or ``Color.reset`` as color causes a reset code to be inserted.

    With respect to :func:`click.style`, this class:

    - has an argument less, ``reset``, which is always ``True``
    - add the ``text_transform``.

    .. warning::
        The arguments ``overline``, ``italic`` and ``strikethrough`` are only
        supported in Click 8 and will be ignored if you are using Click 7.

    :param fg: foreground color
    :param bg: background color
    :param bold:
    :param dim:
    :param underline:
    :param overline:
    :param italic:
    :param blink:
    :param reverse:
    :param strikethrough:
    :param text_transform:
        a generic string transformation; useful to apply functions like ``str.upper``

    .. versionadded:: 0.8.0
    """
    fg: Optional[str] = None
    bg: Optional[str] = None
    bold: Optional[bool] = None
    dim: Optional[bool] = None
    underline: Optional[bool] = None
    overline: Optional[bool] = None
    italic: Optional[bool] = None
    blink: Optional[bool] = None
    reverse: Optional[bool] = None
    strikethrough: Optional[bool] = None
    text_transform: Optional[IStyle] = None

    _style_kwargs: Optional[dict] = dc.field(init=False, default=None)

    def __call__(self, text: str) -> str:
        if self._style_kwargs is None:
            kwargs = dc.asdict(self)
            delete_keys(kwargs, ['text_transform', '_style_kwargs'])
            if int(click_version_tuple[0]) < 8:
                # These arguments are not supported in Click < 8. Ignore them.
                delete_keys(kwargs, ['overline', 'italic', 'strikethrough'])
            object.__setattr__(self, '_style_kwargs', kwargs)
        else:
            kwargs = self._style_kwargs

        if self.text_transform:
            text = self.text_transform(text)
        return click.style(text, **kwargs)


class Color(FrozenSpace):
    """Colors accepted by :class:`Style` and :func:`click.style`."""
    black = "black"
    red = "red"
    green = "green"
    yellow = "yellow"
    blue = "blue"
    magenta = "magenta"
    cyan = "cyan"
    white = "white"
    reset = "reset"
    bright_black = "bright_black"
    bright_red = "bright_red"
    bright_green = "bright_green"
    bright_yellow = "bright_yellow"
    bright_blue = "bright_blue"
    bright_magenta = "bright_magenta"
    bright_cyan = "bright_cyan"
    bright_white = "bright_white"



"""
This example:
- requires click-default-group
- was written in response to a StackOverflow question, not because I think it's
  a good idea to have groups with default command; I tend to not recommend them.
"""
import click

import cloup
from click_default_group import DefaultGroup


class GroupWithDefaultCommand(cloup.Group, DefaultGroup):
    # (Optional) Mark the default command with "*". If you don't need this, you
    # can just write `pass`.
    def format_subcommand_name(
        self, ctx: click.Context, name: str, cmd: click.Command
    ) -> str:
        if name == self.default_cmd_name:
            name = name + "*"
        return super().format_subcommand_name(ctx, name, cmd)


@cloup.group(cls=GroupWithDefaultCommand, default='alice')
def cli():
    pass


@cli.command()
@cloup.option("--foo")
def alice(**kwargs):
    print("Called alice with", kwargs)


@cli.command()
@cloup.option("--bar")
def bob(**kwargs):
    print("Called bob with", kwargs)


if __name__ == '__main__':
    cli()



import click


class Argument(click.Argument):
    def __init__(self, *args, help=None, **attrs):
        super().__init__(*args, **attrs)
        self.help = help

    def help_record(self, ctx):
        return [self.make_metavar(), self.help or ""]


class Option(click.Option):
    """A click.Option with an extra field ``group`` of type ``OptionGroup``."""

    def __init__(self, *args, group=None, **attrs):
        super().__init__(*args, **attrs)
        self.group = group


GroupedOption = Option
"""Alias of ``Option``."""


def argument(*args, help=None, cls=Argument, **kwargs):
    return click.argument(*args, help=help, cls=cls, **kwargs)


def option(*param_decls, cls=None, group=None, **kwargs):
    """Attaches an ``Option`` to the command.
    Refer to :class:`click.Option` and :class:`click.Parameter` for more info
    about the accepted parameters.

    In your IDE, you won't see arguments that has to do with shell completion,
    because they are different in Click 7 and 8 (both supported by Cloup):

    - in Click 7, it's ``autocompletion``
    - in Click 8, it's ``shell_complete``.

    These arguments have different semantics, refer to Click's docs.
    """
    if cls is None:
        cls = Option

    def decorator(f):
        func = click.option(*param_decls, cls=cls, **kwargs)(f)
        new_option = func.__click_params__[-1]
        new_option.group = group
        if group and group.hidden:
            new_option.hidden = True
        return func

    return decorator
